package server;

import java.util.concurrent.locks.ReentrantLock;

/**
 * This class models a printer server.
 * 
 * @author Martijn
 * 
 * @version $Revision: 1.0 $
 */
public class PrinterServer {

	/*@
	 * public resource state(int i) = Perm(queue[i],100) **
	 * ((queue[i] != null) -* queue[i].nodeState());
	@*/
	

	/**
	 * The queue of this printer server.
	 */
	private SimpleListNode[] queue;
	/**
	 * The list of locks where each lock protects an element of the queue.
	 */
	private ReentrantLock[] locks;
	
	/**
	 * Constructs this printer server.
	 * @requires true;
	 * @ensures Perm(queue, 100) ** Perm(length, 100) ** Perm(locks, 100);
	 */
	private PrinterServer() {}
	
	/**
	 * Constructs a PrinterServer with the specified number of users in the queue.
	 * Makes queue, length and locks read-only.
	 * @param numberOfUsers The number of users the printer server supports.
	 * @requires Perm(numberOfUsers, p);
	 * @ensures Perm(queue, 100-p) ** Perm(length, 100-p) ** Perm(locks, 100-p) ** 
	 * (\forall int j; 0 <= j && j < length; Perm(locks[j], 100-p));
	 */
	public PrinterServer(int numberOfUsers) {
		this();
		init(numberOfUsers);
	}
	
	/**
	 * Initializes this printer server. Makes queue, length and locks read-only.
	 * @param numberOfUsers The number of users that this printer server supports.
	 * @requires Perm(queue, 100) ** Perm(locks, 100) ** Perm(numberOfUsers, p);
	 * @ensures Perm(queue, 100-p) ** Perm(locks, 100-p) ** (\forall int j; 0 <= j && j < length; 
	 * Perm(locks[j], 100-p) ** locks[j] != null ** queue[j] != null);
	 */
	private void init(int numberOfUsers) {
		queue = new SimpleListNode[numberOfUsers];
		locks = new ReentrantLock[numberOfUsers];
		
		for(int i = 0; i < queue.length; i++) {
			queue[i] = new SimpleListNode(null);
			/*@ pred inv = state(i); @*/
			locks[i] = new ReentrantLock/*@<inv>@*/();
			//@ locks[i].commit();
		}
	}
	
	/**
	 * Adds a PrintJob to the queue of the specified user.
	 * @param printJob The print job being added.
	 * @param id The ID of the user the print job belongs to.
	 * @requires printJob != null ** id >= 0 ** id < queue.length ** queue[id] != null ** Perm(id, p) ** Perm(locks[id], p) ** Perm(printJob, p);
	 * @ensures queue[id] != null ** Perm(locks[id], p);
	 */
	public void addToQueue(int id, String printJob) {
		locks[id].lock();
		queue[id].add(printJob);
		locks[id].unlock();
	}

	/**
	 * Removes a specific print job from the queue.
	 * @param printJob The print job being removed.
	 * @param id The id of the user the print job belongs to.
	 * @return The print job that was removed from the user's list or <code>null</code> if it wasn't found.
	 * @requires printJob != null ** id >= 0 ** id < queue.length ** Perm(id, p) ** Perm(locks[id], p) ** 
	 * Perm(printJob, p) ** queue[id] != null;
	 * @ensures Perm(locks[id], p) ** queue[id] != null ** Perm(removedElement, 100);
	 */
	public String removeFromQueue(int id, String printJob) {
		String removedElement = null;
		locks[id].lock();
		removedElement = queue[id].remove(printJob);
		locks[id].unlock();
		return removedElement;
	}

	/**
	 * Removes the first print job from the queue.
	 * @param id The id of the user the print job belongs to.
	 * @return The print job that was removed from the user's list or <code>null</code> the list was empty.
	 * @requires id >= 0 ** id < queue.length ** Perm(locks[id], p) ** Perm(id, p) ** queue[id] != null;
	 * @ensures Perm(locks[id], p) ** queue[id] != null ** Perm(removedElement, 100);
	 */
	public String removeFromQueue(int id) {
		String removedElement = null;
		locks[id].lock();
		removedElement = queue[id].remove();
		locks[id].unlock();
		return removedElement;
	}
	
	/**
	 * Removes a user's list of print jobs from the queue.
	 * @param id The id of the user whose print jobs are removed.
	 * @requires id >= 0 ** id < queue.length ** Perm(locks[id], p);
	 * @ensures Perm(locks[id], p);
	 */
	public void removeUser(int id) {
		locks[id].lock();
		queue[id] = new SimpleListNode(null);
		locks[id].unlock();
	}
	
	/**
	 * Retrieves a user's print jobs.
	 * @param id The id of the user whose print jobs are retrieved.
	 * @return The print jobs of the user.
	 * @requires Perm(id, p) ** Perm(locks[id], p);
	 * @ensures Perm(locks[id], p) ** Perm(retrievedJobs, 100);
	 */
	public String[] retrievePrintJobs(int id) {
		locks[id].lock();
		String[] retrievedJobs = (String[]) queue[id].retrieveData().toArray();
		locks[id].unlock();
		return retrievedJobs;
	}
}
